// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file ProblemImplicit.h */

#ifndef AMDIS_PROBLEMIMPLICIT_H
#define AMDIS_PROBLEMIMPLICIT_H

#include "AMDiS.h"
#include "ProblemScal.h"
#include "ProblemVec.h"

namespace AMDiS {
  class ProblemImplicitBase 
  {
  public:
    virtual bool createImplicitMesh() = 0;
    virtual DOFVector< double >* getSignedDistance(int im = 0, int m = 0) = 0;
    virtual DOFVector< double >* getPhi1(int im = 0, int m = 0) = 0;
    virtual DOFVector< double >* getPhi2(int im = 0, int m = 0) = 0;
    virtual DOFVector< double >* getLevelset(int im = 0, int m = 0) = 0;

  protected:
    void readDofVec(std::istream& , DOFVector<double>* , Mesh* );
    void readR(std::istream& , double , Mesh* , int implMesh = 0, int comp = 0);
    void readPhi1(std::istream& , double , Mesh* , int implMesh = 0, int comp = 0);
    void readPhi2(std::istream& , double , Mesh* , int implMesh = 0, int comp = 0);
    std::string getDofFilename(std::string path, int implMesh = 0);
    double getEpsilon(std::string path, int implMesh = 0);
    int getType(std::string path, int implMesh = 0);
    virtual bool isImplicitMesh(int comp = 0 ) = 0;
  };
 
  class ProblemImplicitScal : public ProblemScal , public ProblemImplicitBase 
  {
  public:
    ProblemImplicitScal(std::string name, ProblemIterationInterface* pis = NULL)
      : ProblemScal(name, pis),
        r(0),
        phi1(0),
        phi2(0),
        levelSet(0),
        readImplMesh(false)
        {}

    /// @return: true if implicit mesh type is ok (see below) and false in 
    /// all other cases
    /// initfile entries: 
    /// name+"->implicit mesh->mesh file": name of serialized mesh
    /// name+"->implicit mesh->dof file": name of serialized dofvector
    /// name+"->implicit mesh->type": type of serialized dofvector. 
    /// type can be: 
    ///    0: dofvector is a signed distance r
    ///    1: dofvector is a phasefield phi=0.5*(1-tanh(3*r/eps))
    ///    2: dofvector is a phasefield phi=0.5*(1+tanh(3*r/eps))
    /// assumes:
    ///    a) already created, but not yet initialized mesh
    ///    b) already created feSpace
    virtual bool createImplicitMesh();

    virtual void createMesh();
    virtual void initialize(Flag initFlag, ProblemScal *adoptProblem = NULL,
			    Flag adoptFlag = INIT_NOTHING);
    DOFVector< double >* getSignedDistance(int im = 0 , int m = 0);
    DOFVector< double >* getPhi1(int im = 0, int m = 0); 
    DOFVector< double >* getPhi2(int im = 0, int m = 0);
    DOFVector< double >* getLevelset(int im = 0, int m = 0);

  protected:
    /// DOFVector for a signed distance
    std::vector< DOFVector< double >* > r;

    /// DOFVector for the phasefield function 0.5*(1-tanh(3*r/eps))
    std::vector< DOFVector< double >* > phi1;

    /// DOFVector for the phasefield function 0.5*(1+tanh(3*r/eps))
    std::vector< DOFVector< double >* > phi2;

    /// DOFVector for the levelset function 
    /// (levelSet(x): x \in \Omega: 1, x \not \in Omega: -1, x \in \Gamma: 0)
    std::vector< DOFVector< double >* > levelSet;

    bool readImplMesh;

    bool isImplicitMesh(int comp = 0) { return readImplMesh; }

    bool createImplicitMesh( std::string path, int comp );
  };

  class ProblemImplicitVec : public ProblemVec, public ProblemImplicitBase
  {
  public:
    ProblemImplicitVec(std::string name, 
		    ProblemIterationInterface* problem = NULL)
      : ProblemVec(name, problem),
	r(0),
	phi1(0),
	phi2(0),
	levelSet(0),
        implMesh(0)
    {}

    virtual void createMesh();
    virtual bool createImplicitMesh();
    virtual void initialize(Flag initFlag, ProblemScal* adoptProblem=NULL, 
		            Flag adoptFlag = INIT_NOTHING);
    DOFVector< double >* getSignedDistance(int im = 0, int m = 0) ;
    DOFVector< double >* getPhi1(int im = 0, int m = 0) ; 
    DOFVector< double >* getPhi2(int im = 0, int m = 0) ;
    DOFVector< double >* getLevelset(int im = 0, int m = 0) ;
 
  protected:
    bool createImplicitMesh(int p);
    bool createImplicitMesh(std::string, int, int); 

    /// DOFVector for a signed distance
    std::vector< std::vector< DOFVector< double >* > > r;

    /// DOFVector for the phasefield function 0.5*(1-tanh(3*r/eps))
    std::vector< std::vector< DOFVector< double >* > > phi1;

    /// DOFVector for the phasefield function 0.5*(1+tanh(3*r/eps))
    std::vector< std::vector< DOFVector< double >* > > phi2;

    /// DOFVector for the levelset function 
    /// (levelSet(x): x \in \Omega: 1, x \not \in Omega: -1, x \in \Gamma: 0)
    std::vector< std::vector< DOFVector< double >* > >  levelSet;

    std::vector< bool > implMesh;
    
    bool isImplicitMesh(int comp = 0) { return implMesh[comp]; }
  };
}
#endif
